--
-- BottomArmTrailerAttacher
--
-- @author  Stefan Maurus
-- @date  15/11/2019
--
-- Copyright (C) GIANTS Software GmbH, Confidential, All Rights Reserved.

---Specialization to adjust the bottom arm rotation when attaching a trailer to a certain height and deactivating a suspension node if a tool is attached
-- @category Specializations
BottomArmTrailerAttacher = {}
BottomArmTrailerAttacher.MOD_NAME = g_currentModName

---Checks if all prerequisite specializations are loaded
-- @param table specializations specializations
-- @return boolean hasPrerequisite true if all prerequisite specializations are loaded
-- @includeCode
function BottomArmTrailerAttacher.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(AttacherJoints, specializations) and SpecializationUtil.hasSpecialization(Suspensions, specializations)
end

function BottomArmTrailerAttacher.registerFunctions(vehicleType)
    SpecializationUtil.registerFunction(vehicleType, "setHeightAdjustNodeRotation", BottomArmTrailerAttacher.setHeightAdjustNodeRotation)
end

function BottomArmTrailerAttacher.registerOverwrittenFunctions(vehicleType)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "loadAttacherJointFromXML", BottomArmTrailerAttacher.loadAttacherJointFromXML)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "getIsSuspensionNodeActive", BottomArmTrailerAttacher.getIsSuspensionNodeActive)
end

function BottomArmTrailerAttacher.registerEventListeners(vehicleType)
    SpecializationUtil.registerEventListener(vehicleType, "onLoad", BottomArmTrailerAttacher)
    SpecializationUtil.registerEventListener(vehicleType, "onPreAttachImplement", BottomArmTrailerAttacher)
    SpecializationUtil.registerEventListener(vehicleType, "onPreDetachImplement", BottomArmTrailerAttacher)
end

---
-- @includeCode
function BottomArmTrailerAttacher:onLoad(savegame)
    if self.isClient then
        local i = 0
        while true do
            local key = string.format("vehicle.suspensions.suspension(%d)", i)
            if not hasXMLProperty(self.xmlFile, key) then
                break
            end

            local suspensionNode = self.spec_suspensions.suspensionNodes[i+1]
            if suspensionNode ~= nil then
                suspensionNode.attacherJointIndices = {StringUtil.getVectorFromString(getXMLString(self.xmlFile, key .. "#attacherJointIndices"))}
            end

            i = i + 1
        end
    end
end

function BottomArmTrailerAttacher:loadAttacherJointFromXML(superFunc, attacherJoint, xmlFile, baseName, index)
    if not superFunc(self, attacherJoint, xmlFile, baseName, index) then
        return false
    end

    local node = I3DUtil.indexToObject(self.components, getXMLString(xmlFile, baseName.. ".heighAdjustNode#node"), self.i3dMappings)
    if node ~= nil then
        local heightAdjustNode = {}
        heightAdjustNode.node = node
        heightAdjustNode.rotAxis = getXMLInt(self.xmlFile, baseName.. ".heighAdjustNode#rotAxis")
        heightAdjustNode.detachedRotation = math.rad(getXMLFloat(self.xmlFile, baseName.. ".heighAdjustNode#detachedRotation") or 0)
        heightAdjustNode.ptoBelowRotation = math.rad(getXMLFloat(self.xmlFile, baseName.. ".heighAdjustNode#ptoBelowRotation") or 0)
        heightAdjustNode.rotation = {getRotation(node)}

        heightAdjustNode.curve = AnimCurve:new(linearInterpolator1)
        local i = 0
        while true do
            local baseKey = string.format("%s.heighAdjustNode.key(%d)", baseName, i)
            if not hasXMLProperty(xmlFile, baseKey) then
                break
            end

            local height = getXMLFloat(self.xmlFile, baseKey .. "#height")
            local rotation = math.rad(getXMLFloat(self.xmlFile, baseKey .. "#rotation") or 0)

            heightAdjustNode.curve:addKeyframe({rotation, time = height})

            i = i + 1
        end

        attacherJoint.heightAdjustNode = heightAdjustNode
    end

    return true
end

---
-- @includeCode
function BottomArmTrailerAttacher:onPreAttachImplement(object, inputJointDescIndex, jointDescIndex)
    local spec = self.spec_attacherJoints
    local objectAttacherJoint = object.spec_attachable.inputAttacherJoints[inputJointDescIndex]
    local jointDesc = spec.attacherJoints[jointDescIndex]
    if objectAttacherJoint ~= nil and jointDesc ~= nil then
        local heightAdjustNode = jointDesc.heightAdjustNode
        if heightAdjustNode ~= nil then
            local rotation = heightAdjustNode.curve:get(objectAttacherJoint.attacherHeight)

            if self.getOutputPowerTakeOffsByJointDescIndex ~= nil then
                local ptoOutputs = self:getOutputPowerTakeOffsByJointDescIndex(jointDescIndex)
                if ptoOutputs ~= nil and #ptoOutputs > 0 then
                    local ptoOutput = ptoOutputs[1]
                    local ptoInput = ptoOutput.connectedInput
                    if ptoInput ~= nil then
                        if not ptoInput.aboveAttacher then
                            rotation = heightAdjustNode.ptoBelowRotation
                        end
                    end
                end
            end

            self:setHeightAdjustNodeRotation(heightAdjustNode.node, heightAdjustNode.rotAxis, heightAdjustNode.rotation, rotation)
        end
    end
end

---
-- @includeCode
function BottomArmTrailerAttacher:onPreDetachImplement(implement)
    local spec = self.spec_attacherJoints
    if implement.object ~= nil then
        local jointDesc = spec.attacherJoints[implement.jointDescIndex]
        local heightAdjustNode = jointDesc.heightAdjustNode
        if heightAdjustNode ~= nil then
            self:setHeightAdjustNodeRotation(heightAdjustNode.node, heightAdjustNode.rotAxis, heightAdjustNode.rotation, heightAdjustNode.detachedRotation)
        end
    end
end

---
-- @includeCode
function BottomArmTrailerAttacher:setHeightAdjustNodeRotation(node, rotAxis, baseRotation, rotation)
    local x, y, z = getRotation(node)
    baseRotation[1] = x
    baseRotation[2] = y
    baseRotation[3] = z

    baseRotation[rotAxis] = rotation

    setRotation(node, unpack(baseRotation))

    if self.setMovingToolDirty ~= nil then
        self:setMovingToolDirty(node)
        self:updateCylinderedInitial(false, true)
    end
end

function BottomArmTrailerAttacher:getIsSuspensionNodeActive(superFunc, suspensionNode)
    if not superFunc(self, suspensionNode) then
        return false
    end

    if suspensionNode.attacherJointIndices ~= nil then
        for i=1, #suspensionNode.attacherJointIndices do
            if self:getImplementFromAttacherJointIndex(suspensionNode.attacherJointIndices[i]) ~= nil then
                return false
            end
        end

    end

    return true
end
